#!/usr/bin/env python

import os
import errno
import sqlite3
import json
import time

import requests

"""
This module provide anything needed to create and subclass Box instancs, which
are used to store raw documents (without their metadata.
"""
class NoteEnoughPrivileges(Exception):
    pass

class Box(dict):
    """
    This is an interface to be subclassed by specific Box. It provides standard 
    method that must be present on any of its subclasses and works as a dict.

    It can be used to specialise box. Either using local filesystem, ftp, 
    distributed ones or databases one.

    Subclass of Box must implements _put, _get, _list and _del methods
    """
    def __init__(self, *args, **kwargs):
        """
        Do whatever is needed to have the box Working, like connecting to 
        databases, getting needed parameters, etc
        """
        self.update(*args, **kwargs)

    def __repr__(self, *args, **kwargs):
        """
        Let's display our Box content
        """
        result ='{'
        for (k, v) in self.items(*args, **kwargs):
            result += repr(k) + ": " + repr(v) + ","

        result = result[:-1] + '}'
        return result

    def __len__(self, *args, **kwargs):
        """
        Return the length of the box, ie the number of keys
        """
        return len(self._list(*args, **kwargs))

    def __getitem__(self, key, *args, **kwargs):
        """
        Return the File like object associated to a key
        """
        # And if it exist
        if not self.has_key(key, *args, **kwargs):
            raise IndexError

        return self._get(key, *args, **kwargs)

    def __setitem__(self, key, document, *args, **kwargs):
        """
        Let's assign the document to a specific place in the Box
        """
        self._put(key, document, *args, **kwargs)

    def __delitem__(self, key, *args, **kwargs):
        """
        Delete an item from the box
        """
        self._del(key, *args, **kwargs)

    def __iter__(self, *args, **kwargs):
        """
        Let's iterate over keys
        """
        for key in self.keys(*args, **kwargs):
            yield key

    def __contains__(self, key, *args, **kwargs):
        """
        Do we have a key in our Box
        """
        if key in self._list(*args, **kwargs):
            return True
        return False

    def iterkeys(self, *args, **kwargs):
        """
        Let's iterate over keys
        """
        self.__iter__(*args, **kwargs)

    def itervalues(self, *args, **kwargs):
        """
        Let's iterate over values
        """
        for key in self.iterkeys():
            yield self._get(key, *args, **kwargs)

    def keys(self, *args, **kwargs):
        """
        Returns the ordered file list contained in the box, which are the keys.
        """
        return self._list(*args, **kwargs)

    def values(self, *args, **kwargs):
        """
        Returns the ordered file content in the box, values are File like object.
        """
        return [ self._get(doc, *args, **kwargs) for doc in self.keys(*args, **kwargs) ]

    def items(self, *args, **kwargs):
        """
        Return the tuples (key, value) of the document in the Box
        """
        return [ (key, self._get(key, *args, **kwargs),) for key in self.keys(*args, **kwargs) ]

    def has_key(self, name, *args, **kwargs):
        """
        Check if a document with name exist in the box. Name must be a UUID.
        """
        if not name in self._list(*args, **kwargs):
            return False
        return True

    # Below are the method that needs to be subclassed by specific box systems
    def _list(self):
        """
        Returns the list of all items in the box.
        """
        raise NotImplementedError

    def _get(self, name):
        """
        Get a document from the box, should return a File like object
        """
        raise NotImplementedError

    def _put(self, name, document):
        """
        Put a document in a box. document are, generally, File like objects.

        If a document with this name exist, it will be overwritten.
        """
        raise NotImplementedError

    def _del(self, name):
        """
        Remove a document from the box.
        """
        raise NotImplementedError

class FileBox(Box):
    """
    This box use the filesystem to store document. They're all stored in a 
    directory passed as an arg to __init__
    """
    path = None
    readonly = False

    def __init__(self, path):
        """
        We will test if the path exist, and if not creates it.
        """
        super(FileBox, self).__init__()

        self.path = os.path.abspath(path)
        # We do not have a directory
        try:
            os.mkdir(self.path)
        except IOError as e:
            raise e
        except OSError as e:
            error = errno.errorcode[e.errno]
            if error == 'EEXIST':
                # The file or directory already exist
                if os.path.isfile(self.path):
                    # And it's a file
                    raise e
            else:
                # Something else happened
                raise e

    def _list(self):
        """
        Return a list of files UUID in the FileBox
        """
        return [ unicode(item) for item in os.listdir(self.path)]

    def _put(self, name, document):
        """
        Put a document in the FileBox. name must be an a hash and document
        a File-like object
        """
        # Let's go back to the origi of file
        document.seek(0)

        # Let's open the destination, it will be created if needed
        # And let's write the origin in it by block of 4096bytes
        with open(os.path.join(self.path, unicode(name)), 'wb') as destination, document as origin:
            destination.write(origin.read())

    def _get(self, name):
        """
        Let's retrieve a file from the FileBox. name must be a hash and 
        we will return a File object
        """
        try:
            return open(os.path.join(self.path, unicode(name)), 'rb')
        except:
            raise IOError

    def _del(self, name):
        """
        Let's delete a file by its name
        """
        os.remove(os.path.join(self.path, unicode(name)))

class RemoteBox(Box):
    """
    This box is, in fact, a remote shelf. We play with it by using REST API. But
    since it's a remote one we won't write in it. And so it's a readonly Box.
    """
    readonly = True
    def __init__(self, api_root=""):
        """
        We need a token_id to auth against the remote shelf. And then, to
        auth ourselves, that's somewhere we can save some time.
        """
        super(Box, self).__init__()

        self.api_root = api_root

    def _list(self):
        """
        List does not need auth, which is nice. We need to send a magic query with '*' in it.
        It will returns metadata, we just need to parse to get the media url.

        We must return a list of path.
        """
        # We will query for hash - it's a mandatory field, we sure it exist
        # and we're using the joker '*' because it will give us all the doc
        # stored there.
        data=json.dumps({'hash': '*'})
        headers={'Content-type': 'application/json'}

        results = []

        # Let's create the search.
        r = requests.put(self.api_root+'/search/', data=data, headers=headers)

        try:
            r.raise_for_status()
        except:
            return results

        # The search is in progress, we have to wait for it to be complete
        while r.json()['state'] != "done":
            r = requests.get(self.api_root+r.json()['uri']+'/')
            try:
                r.raise_for_status()
            except:
                return results

        # So now, we have an ended search query, with all the doc in them.
        for hit in r.json()['results']:
            # Each hit is a dict and have a 'doc_url' field, that's what we
            # to query at a later time.
            results.append(hit['doc_url'].split('/')[1])

        return results

    def _put(self):
        """
        We're not going to store a document in a remote-shelf, because we don't have
        the associated metadata.
        """
        raise IOError("Cannot put documents onto a remote shelf")

    def _get(self, name):
        """
        Let's get a doc by its name, the name is a hash, we just need to assemble a URL
        """
        # We do have a document, we need to get the media url. ALso, streaming for the
        # win.
        r = requests.get('/'.join([self.api_root, 'media', name, ''], stream=True)

        # If we have an exception, it means document does not exist or
        # is unavailable
        try:
            r.raise_for_status()
        except:
            raise IOError("No such document in remote shelf")

        # requests .raw attrbute supports file like method (read, etc)
        return r.raw

    def _del(self):
        """
        We're not going to delete document from a remote shelf, it might be used elsewhere.
        """
        raise IOError("Cannot delete document from a remote shelf")



#!/usr/bin/env python

from whoosh.fields import SchemaClass, ID, TEXT, KEYWORD

"""
This module will provides different Schema to be used for Shelves
"""

class BasicSchema(SchemaClass):
    """
    A really basic schema, with only file ID, an associated TEXT and keywords
    """
    hash = ID(unique=True, stored=True) # Name must be unique in a shelves
    description = TEXT(spelling=True, stored=True) # A description of the document being stored
    keywords = KEYWORD(lowercase=True, scorable=True, stored=True) # And associated keywords



#!/usr/bin/env python

import os
from hashlib import sha1

from whoosh import index
from whoosh.query import Every
from whoosh.fields import ID
from whoosh.writing import AsyncWriter
from whoosh.qparser import QueryParser, MultifieldParser

import schemas
import boxes

"""
This module will provide everything needed for shelves to work and expose all
the needed method.

A shelve is made of at least a searchable index, and a readable store (indexed
by the index).
"""

class Shelf(dict):
    """
    This is an interface to be implemented by specific Shelf by subclassing it.
    it provides standard method that musst be present on any of the Shelves.

    boxes is a list of Box which can hold Documents.
    readonly is used to check if you can write/upload/remove/alter content
    index is a whoosh index used to find documents related to a search
    """
    boxes = []
    readonly = False
    index = None

    def __init__(self, *args, **kwargs):
        """
        Do all the necessary initialisation here.
        Mandatory are a Box system and an Index one.
        """
        pass

    def add(self, document, metadata):
        """
        Add a document and it's metadata into a Shelf
        """
        is self.readonly:
            raise IOError("Shelf is readonly")
        raise NotImplementedError

    def delete(self, name):
        """
        Delete a document and it's metadata from a Shelf.
        we use a name to find it, it must be more a serial number
        than a real name.
        """
        is self.readonly:
            raise IOError("Shelf is readonly")
        raise NotImplementedError

    def modify(self, name, metadata):
        """
        Modify a document by its name. Metadata are merged with the previous one.
        The doc isn't changed.
        """
        is self.readonly:
            raise IOError("Shelf is readonly")
        raise NotImplementedError

    def search(self, query):
        """
        Search through the metadata index of a Shelf and return
        metadata including document id
        """
        raise NotImplementedError

    def get(self, name):
        """
        Get the document named name from one of the boxes
        """
        raise NotImplementedError

class FileShelf(Shelf):
    """
    This class implement a shelf with a classic Whoosh index on filesystem and filed with FileBoxes
    """

    def __init__(self, index_dir, schema_class=schemas.BasicSchema, *args, **kwargs):
        """
        Initialize the shelf

        index_dir is where the Index will be stored
        schema_class is a class used as the schema of the shelf.
        """
        super(WhooshFileShelf, self).__init__(*args, **kwargs)

        # If the index_dir does not exists, let's create it
        if not os.path.exists(index_dir):
            os.mkdir(index_dir)

        # If there's already an index in index_dir, just open it
        if index.exists_in(index_dir):
            self.index = index.open_dir(index_dir)
        else:
            self.index = index.create_in(index_dir, schema_class())


    def __update_index(self, metadata):
        """
        This method is used inetrnally to update the index when needed
        """
        # Let's grab a writer to the index
        writer = AsyncWriter(self.index)

        # And now, let's crawl the metadata for unknown fields
        known_fields = self.index.schema.names()
        for k in metadata.keys():
            if k not in known_fields:
                writer.add_field(k, TEXT(stored=True))

        # We just need to add the document to the index now
        writer.update_document(**metadata)

        # Commit and close
        writer.commit()

    def add(self, document, metadata={}):
        """
        Add a document to the file index and to the boxes. Metadata is a dict.
        Each field not present in the current schema is added to it.
        """
        # We need at least one box on our shelf to store documents
        if len(self.boxes) == 0:
            raise IndexError

        # We generate a name for the document. The name must be the same for two similar files
        # So we need to hash it first
        shash = sha1()
        with open(document) as f:
            shash.update(f.read())
        name = unicode(shash.hexdigest())

        metadata['hash'] = name

        # Write everything into the index
        self.__update_index(metadata)

        # Let's check if the document exist in a FileBox
        exist = False
        for box in self.boxes:
            if box.has_key(unicode(name)):
                exist = True
                break

        # if it does not, add it to the first writable Box
        for box in self.boxes:
            if box.readable:
                box[0][unicode(name)] = open(document)
                break

        # We return the name of the newly created doc
        return name

    def delete(self, name, purge=True):
        """
        Delete a document by its name. name is actually a hash. If purge is true, file is also
        removed from the boxes.
        """
        # Grab a writer on the index
        writer = AsyncWriter(self.index)

        # Delete and commit ffom index
        writer.delete_by_term(u'hash', name)
        writer.commit()

        # Delete the document from the boxes if we want to purge them
        if not purge:
            return

        # We need to remove the doc is box is writable
        for box in self.boxes:
            if box.haskey(name) and not box.readonly:
                del(box[name])

    def modify(self, name, metadata):
        """
        Modify a document metadata
        """
        # Let's check if the document exist
        # It must be present in the index
        qp = QueryParser("hash", schema = self.index.schema)
        q = qp.parse(name)
        with self.index.searcher() as s:
            # No results
            if len(s.search(q)) == 0:
                raise IOError("Document does not exist")

        # So we do have a document, we just need to update it
        metadata['hash'] = name

        # Write everything into the index
        self.__update_index(metadata)

    def search(self, query):
        """
        Let's send a query to the shelf. The query is a dict of keys, with associated values.
        """
        qp = MultifieldParser(query.keys(), schema=self.index.schema)

        # Let's assemble the query
        search_terms = ''
        for k in query:
            search_terms += '%s:%s ' % (k, query[k])

        # We need to parse the query
        if u'*' in [ query[k] for k in query]:
            # We have a query asking for every doc
            q = Every()
        else:
            q = qp.parse(search_terms)

        # And now, we search
        results = self.index.searcher().search(q, limit=None)

        # We just have to return the results
        return results

    def get(self, name):
        """
        Let's return File like object from their names
        """
        for box in self.boxes:
            if name in box.keys():
                return box[name]
        return None

def ShelfOfShelves(Shelf)
    """
    This Shelf is made of shelves. We will thread and assemble all queries. The shelf of shelves
    is readonly.

    library is the list of shelves.
    """
    readonly = True
    library = []

    def __init__(self, shelves_list=[]):
        """
        Let's initialize our library.

        shelves_list is a list of already created shelves that we wanna add
        """
        for shelf in shelves_list:
            self.__add_shelg(self, shelf)

    def __add_shelf(self, shelf):
        """
        Let's add a new shelf to our library.

        shelf must be a Shelf object
        """
        # We will check that the shelf does not exist already in our library
        if shelf not in self.shelves:
            # We must be sure that our shelf is indeed a Shelf
            if not isinstance(shelf, Shelf):
                raise TypeError("%s not a subclass of Shelf" % (type(shelf),))
            library.append(shelf)

            # We will add all the boxes in the same shelf, for ease of access
            for box in shelf.boxes:
                self.boxes.append(box)

    def get(self, name):
        """
        Let's search for a specific document and returns it

        name is the name of the document we're looking for
        """
        for box in self.boxes:
            if name in box.keys():
                return box[name]
        return None

    def search(self, query):
        """
        Let's search through all the shelves and extend the results to include all of them before
        returning it.

        query is a dict of keys/values to search for.
        """
        # SO, let's create our results
        results = None
        for shelf in self.library:
            # First shelf we're running the search on
            if results is None:
                results = shelf.search(query)
            # We use upgrade_and_extend, because multiple hits can happen.
            else:
                results.upgrade_and_extend(shelf.search(query))

        return results



#!/usr/bin/env python
import uuid
import os
from boxes import Safe, FileBox

if __name__ == "__main__":
    box = Safe(FileBox('test'))
    for item in os.listdir('.'):
        if os.path.isfile(item):
            name = uuid.uuid4()
            box[name] = open(item)

    print box
    print box.items()
    print box.keys()
    del(box[box.keys()[1]])
    print box.items()



#!/usr/bin/env python

from distutils.core import setup

setup(name = 'Shelves'
    , version='0.3.1'
    , description='Shelves to store Box of documents in the brodacast-system'
    , author='Okhin'
    , author_email='okhin@okhin.fr'
    , url='https://git.leloop.org/broadcast-system/shelves'
    , packages = ['shelves']
    , requires = ['tahoe_whoosh', 'whoosh', 'requests']
    )
